#ifndef __ICMP_TUNNEL_HDR
#define __ICMP_TUNNEL_HDR

/*
** Includes, defines and global variables used between functions.
*/

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>
extern int errno;

#include <sys/types.h>
#include <sys/time.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#ifdef linux
  #include "linux_ip_icmp.h"
#else
  #include <netinet/ip_icmp.h>
  #include <netinet/ip.h>
#endif
#include <arpa/inet.h>
#include <netdb.h>

#include "error.h"


#ifndef INADDR_NONE
#define INADDR_NONE   0xFFFFFFFF    /* should be in <netinet/in.h> */
#endif

#define SIZE_ICMP_HDR        8      /* 8-byte ICMP header */
#define MAXMESG           4096      /* max data area */
#define MAXPACKET         5004      /* max packet size */
                                    /* SIZE_ICMP_HDR + MAXMESG */
                                    
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN      64      /* should be defined in <sys/param.h> */
#endif

/*
** Our errno stuff.
*/
extern int  eit_errno;

#define EIT_2BIG            0   /* The message we want to sent is too big to */
                                /* fit into 1 packet */
#define EIT_SENDTO          2   /* sendto() error */
#define EIT_SENDLESS        3   /* sendto() send less bytes as it should */
#define EIT_RECVFROM        5   /* recvfrom() error */


struct sockaddr_in          clisrc;

/*
** Our library functions.
*/
void  tunnel_init(void);
int   tunnel_send(char *send_mesg, size_t mesglen, u_long srv_ip);
int   tunnel_recv(char *recv_mesg, size_t mesglen);
void  tunnel_reset(void);

#endif
