/****************************************************************************
**
** Name:	searcher
**
** Usage:	searcher [-a] [-r] [-n] [-f] [-m <word-list>] [-i] [-v]
**			 [-g <group name>] [-s <word-list>] [-l <word-list>]
**		-a: shows all files in the home-directories. (not recommended)
**		-r: lists infos about all files _beginning_ w/ ".rhosts"
**		-n: lists infos about all files _beginning_ w/ ".netrc"
**		-f: lists infos about all files _beginning_ w/ ".forward"
**		-m: scans the users' mbox-file/M|mail-dir./Spool-file for words
**		    defined in <word-list>
**		-i: if given, searcher will ignore locked users (see LOCK_IND)
**		-v: verbose; print a bit more (useless) infos
**		    If this option is given twice and the -m option is also
**		    given, then the lines, which match the pattern, will be
**		    printed to stdout. Yeah, and if "-v" is given 3+ times ALL
**		    lines will be printed.
**		    All non-printable (i.e. PGP Mail) characters will be shown
**		    as decimal number included in brackets "[]".
**		-g: search only in the $HOMEs of users in group <group>
**		-s: super search: serach the words specified in <word-list>
**		                  in ALL files located in $HOME
**		-l: log-search:	scan the log files given in SYSLOG
**			        for words defined in <word-list>
**				ALL log files will be examined, not only the
**				log file for "auth.*".
**				BUT don't forget the "loghost", e-mail,
**				"/dev/console", etc.
**				This prog. can't hack the "loghost" for you,
**				can't blow up the monitor and I'm too lazy to
**				scan the e-mail (now), ;) it's your turn.
**				Hm, but it's a good idea. I will try to solve
**				this "problem"... later.
**
** Notes:		! ALL comparisions are case _in_sensitive. !
**		     ! ALL words in the wordlists must end with a <CR> !
**			   ! Searcher expect one word per line !
**                     ! The word should not be longer than 80 chars !
**
** Purpose:	Scan users' $HOME for specified files and print their
**		status/contents.
**
** Disadvantages: Very CPU/Time consuming.
**		  If NFS is used, then the program produce heavy net-load.
**
******************************************************************************
**
** Version:	8.0
** Last Update:	1997/06/24
**
** Author:	TICK (t_i_c_k@hotmail.com)
**
******************************************************************************
**
** Tested on:	Linux 1.1.59 w/o NFS/NIS
**		Linux 2.0 w/ NFS/NIS
**    Linux 2.0.33 w/o NFS/NIS
**		AIX 4.* w/ NFS/NIS (NAME_MAX is not predefined)
**		AIX 3.2 w/ NFS/NIS (you must include stdlib.h for getopt)
**
*****************************************************************************/

#include <stdio.h>
#include <ftw.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <time.h>
#include <utime.h>

#define RHOSTS		".rhosts"
#define NETRC		".netrc"
#define FORWARD		".forward"
#define MBOX		"mbox"
#define MAILDIR		"/mail/"
#define SPOOL		"/var/spool/mail/"
#define SYSLOG		"/etc/syslog.conf"

#define SCAN		1
#define PRINT		2
#define LOCK_IND	'*'	/* defines the char, which indicates a locked
				   passwd entry */
#define COMMENT		'#'	/* char. used to indicate a comment in SYSLOG */
#define DEPTH		20	/* depth of direc.tree to go down */
#define PROG		"emacs"	/* programname to show in the process list */
#ifndef NAME_MAX
  #define NAME_MAX	255
#endif
#define STRMAX		PATH_MAX+NAME_MAX

#define usage(prog)	\
{	\
	fprintf(stderr, "\nUsage:\t%s\t[-a] [-r] [-n] [-f] [-m <word-list>] [-i] [-v]\n\t\t\t[-g <group name>] [-s <word-list>] [-l <word-list>]\n\n", prog);	\
	exit(255);	\
}


/*****************************************************************************
		G L O B A L - D E C L A R A T I O N S
*****************************************************************************/

int	iSearch4Files(struct passwd *stPwd);
int	iChecker(const char *cFile, struct stat *stStatus, int iFlag);
int	iFindTarget(void);
int	iExamineLogs(void);
void	vCheckFile(const char *cFilePath, struct stat *stStatus, int iFlag);
void	vSetUpWordList(char *(**cArray), char *cFile);
void	vScanFile(char **cWL, FILE *file, const char *cFilePath);
void	vFindLogs(char *(**cArray));
char	*cStr2Lower(char *cString);

unsigned int	uiA, uiR, uiN, uiF, uiM, uiI, uiV, uiG, uiS, uiL;
char		cFile_Mail[PATH_MAX+NAME_MAX+1];
char            cFile_Super[PATH_MAX+NAME_MAX+1];
char            cFile_Logs[PATH_MAX+NAME_MAX+1];
char		**cWL_Mail;
char            **cWL_Super;
char            **cWL_Logs;
char    	cGroup[9];


/*****************************************************************************

				M A I N

******************************************************************************/				

int main(int argc, char **argv)
{
	int	iOpt, iCtr;
	char	cProgname[PATH_MAX+NAME_MAX+1];

	setuid(0);
	if(getuid())
	{
		fprintf(stderr, "\n%s must run by root!\n\n", argv[0]);
		exit(100);
	}
	
 	if(argc == 1)
 		usage(argv[0])
 	 
	strcpy(cProgname, argv[0]);
	/* fake ps list */
	for(iCtr = 0; argv[0][iCtr] != '\0'; iCtr++)
		argv[0][iCtr] = 0;
	strcpy(argv[0], PROG);
	/* done */

	iOpt = uiA = uiR = uiN = uiF = uiM = uiI = uiV = uiG = uiS = uiL = 0;
	opterr = 0;
	while( (iOpt = getopt(argc, argv, "arnfm:ivg:s:l:")) != EOF)
 	{
		switch(iOpt)
	        {
	        	case 'a':
	        		uiA++;
	                        break;
			case 'r':
				uiR++;
				break;
			case 'n':
				uiN++;
				break;
			case 'f':
				uiF++;
				break;
			case 'm':
				uiM++;
				strncpy(cFile_Mail, optarg, sizeof(cFile_Mail)-1);
				break;
			case 'i':
				uiI++;
				break;
			case 'v':
				uiV++;
				break;
			case 'g':
				uiG++;
				strncpy(cGroup, optarg, sizeof(cGroup)-1);
				if(getgrnam(cGroup) == NULL)
				{
					fprintf(stderr, "\nDon't know group: %s\n\n", cGroup);
					exit(1);
				}
				break;
			case 's':
				uiS++;
				strncpy(cFile_Super, optarg, sizeof(cFile_Super)-1);
				break;
			case 'l':
				uiL++;
				strncpy(cFile_Logs, optarg, sizeof(cFile_Logs)-1);
				break;
			case '?':
			default:
				usage(cProgname)
		}
	}
	
	if(uiG && !(uiA + uiR + uiN + uiF + uiM + uiS)) /* only -g is given */
	{
		fprintf(stderr, "\nFor which file should I looking for?\n\n");
		exit(-1);
	}
	
	/* Set up wordlists */
	if(uiM)
		vSetUpWordList(&cWL_Mail, cFile_Mail);
	if(uiS)
		vSetUpWordList(&cWL_Super, cFile_Super);
	if(uiL)
		vSetUpWordList(&cWL_Logs, cFile_Logs);
		
	if(uiF || uiN || uiR || uiA || uiM || uiS)
		exit(iFindTarget());
	if(uiL)
		exit(iExamineLogs());
}


/****************************************************************************
			S U B   R O U T I N E S
****************************************************************************/

/*
** vSetUpWordList stores the words defined in cFile in the array **cArray,
** the words will be stored in lowercase
*/

void vSetUpWordList(char *(**cArray), char *cFile)
{
	FILE	*Fd;
	char	cBuffer[81];
	char	*cWord;
	int	iCtr;
	
	
	if((Fd = fopen(cFile, "r")) == NULL)
	{
		fprintf(stderr, "\nCan't open file %s for reading.\n\n", cFile);
		exit(-3);
	}
	
	*cArray = NULL;
	iCtr = 0;
	while(fgets(cBuffer, sizeof(cBuffer), Fd) != NULL)
	{
		cWord = cStr2Lower(cBuffer);
		/* remove the '\n' */
		cWord[strlen(cWord)-1] = '\0';
		
		if(((*cArray) = (char **) realloc((*cArray), (iCtr+1) * sizeof(char *))) == NULL
		    || ((*cArray)[iCtr] = (char *) malloc(strlen(cWord) + 1)) == NULL)
		{
			fprintf(stderr, "\nInsufficient memory.\n\n");
			exit(-4);
		}
		strcpy((*cArray)[iCtr], cWord);
		
		iCtr++;
	}
	if(ferror(Fd))
	{
		fprintf(stderr, "\nError while reading file: %s\n\n", cFile);
		exit(-5);
	}
	
	/* Set End-of-Array Mark */
	if(((*cArray) = (char **) realloc((*cArray), (iCtr+1) * sizeof(char *))) == NULL
            || ((*cArray)[iCtr] = (char *) malloc(sizeof(NULL))) == NULL)
	{
		fprintf(stderr, "\nInsufficient memory.\n\n");
	        exit(-4);
	}
	(*cArray)[iCtr] = NULL;
	
	fclose(Fd);
	return;
}	


/*
** cStr2Lower converts a string with uppercase letters to a string with
** lowercase letters. The new string will be returned. If an error occurs
** NULL will be returned.
** cString should not be longer than STRMAX bytes.
*/

char *cStr2Lower(char *cString)
{
	char	cConvStr[STRMAX+1];
	int	iCtr;
	
	/* truncate cString */
	if(strlen(cString) > (STRMAX))
		cString[STRMAX] = '\0';
		
	memset(cConvStr, 0, sizeof(cConvStr));
	for(iCtr = 0; cString[iCtr] != '\0'; iCtr++)
		cConvStr[iCtr] = (char) tolower((int) cString[iCtr]);
	
	return(cConvStr);
}



/*
** looksup pwd-entries for users and calls iSearch4Files to search for
** the specified files.
*/

int iFindTarget(void)
{
	struct passwd   *stPwd;
	struct group    *stGrp;
	uid_t	tReUid;
	gid_t	tEGid;
	unsigned int	uiCtr;
                
	tReUid = getuid();
	tEGid = getegid();
	
	if(uiG)
	{
		if((stGrp = getgrnam(cGroup)) == NULL)
		{
			fprintf(stderr, "\nERROR: getgrnam()\n\n");
			return(2);
		}
		
		for(uiCtr = 0; stGrp->gr_mem[uiCtr] != NULL; uiCtr++)
		{
			if((stPwd = getpwnam(stGrp->gr_mem[uiCtr])) == NULL)
			{
				fprintf(stderr, "\nNo entry for user: %s.\n\n", stGrp->gr_mem[uiCtr]);
				continue;
			}
			
			/* change UID to the user's UID */
                        if(stPwd->pw_uid)
                        {
                        	setegid(stPwd->pw_gid);
                        	setreuid(0, stPwd->pw_uid);
                        }
                        /* done */

                        /* START SCANNING */
                        iSearch4Files(stPwd);

                        /* set UID to 0 */
                        if(stPwd->pw_uid)
                        {
                        	setuid(0);
                        	setreuid(tReUid, 0);
                        	setegid(tEGid);
                        }
                        /* done */
		}
	}
	else
	{
		setpwent();
	        while( (stPwd = getpwent()) != NULL)
	        {
			/* change UID to the user's UID */
			if(stPwd->pw_uid)
			{
				setegid(stPwd->pw_gid);
				setreuid(0, stPwd->pw_uid);
			}
			/* done */
		
			/* START SCANNING */
			iSearch4Files(stPwd);
		
			/* set UID to 0 */
			if(stPwd->pw_uid)
			{
				setuid(0);
				setreuid(tReUid, 0);
				setegid(tEGid);
			}
			/* done */
		                                        
		}
		endpwent();
	}
	printf("\n\n");	
	
	return(0);
}



/*
** iSearch4Files walks down the direc.tree of the user's $HOME and calls
** iChecker to check the files.
*/

int iSearch4Files(struct passwd *stPwd)
{
	struct stat	stStatus;
	char		cSpoolFile[PATH_MAX+NAME_MAX+1];

	/* ignore locked accounts ? */
	if(stPwd->pw_passwd[0] == LOCK_IND && uiI)
	{
		if(uiV)
			fprintf(stdout, "\nLOCKED:\tUSER: %s", stPwd->pw_name);
		return(1);	
	}
        
	fprintf(stdout, "\nSEARCH:\tUSER: %s\t\tDIR: %s", stPwd->pw_name, stPwd->pw_dir);
    
	/* walk down the $HOME-direc.tree */    
	if( ftw(stPwd->pw_dir, iChecker, DEPTH) < 0)
		fprintf(stderr, "\nERROR: ftw()\n\n");
    	
    	if(uiM)		/* Mail-Spool-Direc. */
    	{
    		strcpy(cSpoolFile, SPOOL);
    		strcat(cSpoolFile, stPwd->pw_name);
    		
    		if(!access(cSpoolFile, R_OK))
    		{
    			if(stat(cSpoolFile, &stStatus) == -1)
	    		{
    				fprintf(stderr, "\nCan't get status of file %s\n\n", cSpoolFile);
    				return(-1);
    			}
	    		if(stStatus.st_size > 0)
    				vCheckFile(cSpoolFile, &stStatus, SCAN);
    		}
    	}

	return(0);
}



/*
** iChecker checks for the defined files
*/

int iChecker(const char *cPath, struct stat *stStatus, int iFlag)
{
	char *cFile;
	char *cCmpFile;
	char *cCmpDir;
  
	if( (cFile = strrchr(cPath, '/')) != NULL)
		cFile++;
	
	if(uiA)
		printf("\n\t\t%s", cPath);


	cCmpFile = NULL;
	if((cCmpFile = cStr2Lower(cFile)) == NULL)
	{
		if(uiV)
	        	fprintf(stdout, "\nERROR: cStr2Lower()\n\n");
	        return(0);
	}
	
	if(cCmpFile[0] == '.' && (uiF || uiN || uiR) && iFlag == FTW_F)
	{
		if(!strncmp(cCmpFile, RHOSTS, strlen(RHOSTS)) && uiR)
			vCheckFile(cPath, stStatus, PRINT);
		else if(!strncmp(cCmpFile, NETRC, strlen(NETRC)) && uiN)
                        vCheckFile(cPath, stStatus, PRINT);
		else if(!strncmp(cCmpFile, FORWARD, strlen(FORWARD)) && uiF)
			vCheckFile(cPath, stStatus, PRINT);
	}
	if(uiM && iFlag == FTW_F)
	{
		cCmpDir = cStr2Lower((char *) cPath);
		
		/* Check mbox */
		if(!strncmp(cCmpFile, MBOX, strlen(MBOX)))
			vCheckFile(cPath, stStatus, SCAN);
		
		/* Check the Mail-Direc. */
		if(strstr(cCmpDir, MAILDIR) != NULL && iFlag == FTW_F)
			vCheckFile(cPath, stStatus, SCAN);
	}
	if(uiS && iFlag == FTW_F)
		vCheckFile(cPath, stStatus, SCAN);		
		    
	return(0);
}    



/*
** vCheckFile print the file status/contents to stdout
** and also checks the Mail for given words
*/

void vCheckFile(const char *cFilePath, struct stat *stStatus, int iFlag)
{
	FILE		*file;
	char		cPath[PATH_MAX+NAME_MAX+1];
	char		cLine[81];
	struct utimbuf	stRestoreTime;
  
	
	if( (file = fopen(cFilePath, "r")) == NULL)
	{
		printf("\t\tCan't open %s for reading!", cFilePath);
		return;
	}
	
	
	printf("\n\tFILE-PATH:\t%s", cFilePath);
		
	if(uiV)
	{
		printf("\n\tFILE STATUS:");
		printf("\n\t\tLAST ACCESSTIME:\t%s", asctime(localtime(&stStatus->st_atime)));
		printf("\t\tLAST MODIFIKATIONTIME:\t%s", asctime(localtime(&stStatus->st_mtime)));
		printf("\t\tLAST INODEMOD.TIME:\t%s", asctime(localtime(&stStatus->st_ctime)));
		printf("\t\tSIZE: %lu BYTES", stStatus->st_size);
		printf("\n\t\tPROTECTION: %o", stStatus->st_mode);
	}

	printf("\n");
	if(iFlag == PRINT && (uiF || uiN || uiR))
	{
		printf("\n\tFILE CONTENTS:\n");
		while( fgets(cLine, sizeof(cLine), file) != NULL)
			printf("\t\t%s", cLine);
      	}
	if(iFlag == SCAN)
	{
		if(uiM)
		{
			printf("\tMAIL-SCAN\n");
			vScanFile(cWL_Mail, file, cFilePath);
		}
		if(uiS)
		{
			printf("\tSUPER-SCAN\n");
			vScanFile(cWL_Super, file, cFilePath);
		}
		if(uiL)
		{
			printf("\tLOG-SCAN\n");
			vScanFile(cWL_Logs, file, cFilePath);
		}
	}  
	printf("\n");

	/* hooo .... don't leave a fingerprint ;;;;;)))))))) */
	stRestoreTime.actime = stStatus->st_atime;
	stRestoreTime.modtime = stStatus->st_mtime;
	utime(cFilePath, &stRestoreTime);
  
	fclose(file);
	return;
}

/*
** Scan files contents for defined words
*/

void vScanFile(char **cWL, FILE *file, const char *cFilePath)
{
	char	cBuffer[81];
	char	*cLowerCase;
	int	iCtr, iCtr2;

	rewind(file);
	while(fgets(cBuffer, sizeof(cBuffer), file) != NULL)
        {
		cLowerCase = cStr2Lower(cBuffer);
		
		if(uiV >= 3 && uiM)
			for(iCtr = 0; cBuffer[iCtr] != '\0'; iCtr++)
		 		if(isprint(cBuffer[iCtr]) || cBuffer[iCtr] == '\n' || cBuffer[iCtr] == '\t')
                			printf("%c", cBuffer[iCtr]);
                		else
                			printf("[%d]", cBuffer[iCtr]);
                        
		for(iCtr = 0; cWL[iCtr] != NULL; iCtr++)
			if(strstr(cLowerCase, cWL[iCtr]) != NULL)
			{
				if(uiV == 2  && uiM)
					for(iCtr2 = 0; cBuffer[iCtr2] != '\0'; iCtr2++)
						if(isprint(cBuffer[iCtr2]) || cBuffer[iCtr2] == '\n' || cBuffer[iCtr2] == '\t')
				                	printf("%c", cBuffer[iCtr2]);
						else
				                	printf("[%d]", cBuffer[iCtr2]);
				
				printf("\t\tFOUND: %s\n", cWL[iCtr]);
			}
	}
	if(ferror(file))
		fprintf(stderr, "\nError while reading %s\n", cFilePath);
}

/*
** Scan "/etc/syslog.conf" for log files and store them in an 2D-Array
*/

void vFindLogs(char *(**cArray))
{
	char	*cLogFile;
	char	cBuffer[1024];
	FILE	*file;
	int	iCtr, iCtr2;
	struct stat	stStatus;
	
	if((file = fopen(SYSLOG, "r")) == NULL)
	{
		fprintf(stderr, "\nCan't open file %s for reading.\n\n", SYSLOG);
		exit(-3);
	}
	
	*cArray = NULL;
	iCtr = 0;
	while(fgets(cBuffer, sizeof(cBuffer), file) != NULL)
	{
		if(cBuffer[0] == COMMENT)
			continue;
			
		if((cLogFile = strchr(cBuffer, '/')) == NULL)
			continue;
		
		/*printf("%s", cLogFile);*/
		for(iCtr2 = 0; cLogFile[iCtr2] != '\n' && cLogFile[iCtr2] != ','; iCtr2++)
			;
		cLogFile[iCtr2] = '\0';
		
                if(stat(cLogFile, &stStatus) == -1)
                {
                	fprintf(stderr, "\nCan't get status of file: %s.\n", cLogFile);
			continue;
		}
		
		if(!S_ISREG(stStatus.st_mode))      	
                	continue;
                
                if(((*cArray) = (char **) realloc((*cArray), (iCtr+1) * sizeof(char *))) == NULL
                || ((*cArray)[iCtr] = (char *) malloc(strlen(cLogFile) + 1)) == NULL)
		{
                	fprintf(stderr, "\nInsufficient memory.\n\n");
                        exit(-4);
		}
                strcpy((*cArray)[iCtr], cLogFile);

                iCtr++;
	}

	if(ferror(file))
	{
                fprintf(stderr, "\nError while reading file: %s\n\n", SYSLOG);
                exit(-5);
	}
	
	
        /* Set End-of-Array Mark */
        if(((*cArray) = (char **) realloc((*cArray), (iCtr+1) * sizeof(char *))) == NULL
        || ((*cArray)[iCtr] = (char *) malloc(sizeof(NULL))) == NULL)
	{
        	fprintf(stderr, "\nInsufficient memory.\n\n");
                exit(-4);
	}
        (*cArray)[iCtr] = NULL;

        fclose(file);
}

/*
** Check the log files
*/

int iExamineLogs(void)
{
	char	**cLogs;
	int	iCtr;
	struct stat	stStatus;
	
	
	/* find the log files */
	vFindLogs(&cLogs);
	
	/* scan log file */
	for(iCtr = 0; cLogs[iCtr] != NULL; iCtr++)
	{
		if(stat(cLogs[iCtr], &stStatus) == -1)
			continue;
			
		vCheckFile(cLogs[iCtr], &stStatus, SCAN);
	}
	
	return(0);
}