/*
** StrHide is designed to hide Strings in Programs, so Admins can't get
** Informations about your Prog. by just using 'strings'.
** To Hide Things StrHide just xor's an String and then converts the Result
** to Radix64 (Algorithm stolen from uu-en/de-code).
**
** Last Update: 1998/01/27
** Author:      TICK
*/

#include <stdio.h>
#include <sys/types.h>
#include "strhide.h"

#define XOR         81
#define BUFSIZE   4096

#define ENC(c)  ((c) ? ((c) & 077) + ' ': '`')
#define DEC(c)  (((c) - ' ') & 077)


char *encode(char *xorstr, u_int len)
{
  register int  ch, n;
  register char *p, *pe;
  static   char encstr[BUFSIZ] = { 0 };
  
  
  if(len > BUFSIZ/2)
    return(NULL);
  
  pe = encstr;
  *pe++ = ENC(len);
  for(p = xorstr, n = len; n > 0; n -= 3, p += 3)
  {
    ch    = *p >> 2;
    ch    = ENC(ch);
    *pe++ = ch;
    
    ch    = ((*p << 4) & 060) | ((p[1] >> 4) & 017);
    ch    = ENC(ch);
    *pe++ = ch;
    
    ch    = ((p[1] << 2) & 074) | ((p[2] >> 6) & 03);
    ch    = ENC(ch);
    *pe++ = ch;
    
    ch    = p[2] & 077;
    ch    = ENC(ch);
    *pe++ = ch;
  }
  
  *pe = 0;
  
  return(encstr);
}


char *decode(char *encstr)
{
  register int  n;
  register char ch, *p, *pe;
  static   char decstr[BUFSIZ];
  

  for(n = DEC(*encstr++), p = encstr, pe = decstr; n > 0; p += 4, n -= 3)
  {
    if(n >= 3)
    {
      ch    = DEC(p[0]) << 2 | DEC (p[1]) >> 4;
      *pe++ = ch;
      
      ch    = DEC(p[1]) << 4 | DEC (p[2]) >> 2;
      *pe++ = ch;

      ch    = DEC(p[2]) << 6 | DEC (p[3]);
      *pe++ = ch;
    }
    else
    {
      if(n >= 1)
      {
        ch    = DEC(p[0]) << 2 | DEC (p[1]) >> 4;
        *pe++ = ch;
      }
      if(n >= 2)
      {
        ch    = DEC(p[1]) << 4 | DEC (p[2]) >> 2;
        *pe++ = ch;
      }
    }
  }

  *pe = 0;
  
  return(decstr);
}

/*
** str:     plain text or encrypted text (MUST BE 0-terminated)
** mode:    ENCODE = encryption
**          DECODE = decryption
** return:  plain/encrypted text
*/
char *StrHide(char *str, u_int mode)
{
  static char   buffer[BUFSIZ] = { 0 };
  char          *ptr;
  u_int         ctr;
  

  bzero(buffer, sizeof(buffer));
  switch(mode)
  {
    case ENCODE:
      for(ctr = 0; str[ctr] && ctr < sizeof(buffer)-1; ctr++)
        buffer[ctr] = str[ctr] ^ XOR;

      return(encode(buffer, ctr));

    case DECODE:
      ptr = decode(str);

      for(ctr = 0; ptr[ctr] && ctr < sizeof(buffer)-1; ctr++)
        buffer[ctr] = ptr[ctr] ^ XOR;

      return(buffer);

    default:
      return(NULL);
  }
}
