/* Patched 1997 by van Hauser / THC - you are allowed to distribute it

   Using finger for file transfers - can be used for hacking stuff, like
   doing "finger marryv11.c@yourdialup.host.com > marryv11.c"
   to transfer a file from your dialuphost to your current host - but
   i'm sure you can legal stuff with it too ;-)

   If you define USE_FILE_SERVER then you don't need to put the -s option
   to fingerd in inetd.conf
   if you define NO_USER then the normal finger service is disabled.
   You need to define FILE_DIR and point it to the directory where the
   files are placed. NOTE! everyuser must have access to the files in
   FILE_DIR otherwise it needs suid or gid rights!
   FILE_DIR also must end with a / like #define FILE_DIR "/prg/etc/export/"

   Btw, no backdoors put in, no buffer overflow possible.
   If the user tries something like "finger ../../etc/passwd@yourhost.com"
   it will be detected, logged and discarded.

   Have fun. - van Hauser / THC
*/
#define FILE_DIR "/prg/etc/export/"
#define USE_FILE_SERVER
#define NO_USER


char copyright[] =
  "@(#) Copyright (c) 1983 The Regents of the University of California.\n"
  "All rights reserved.\n";
 /*
  * from: @(#)fingerd.c  5.6 (Berkeley) 6/1/90"
  */
char rcsid[] = "$Id: fingerd.c,v 1.9 1996/08/16 21:58:17 dholland Exp $";
        

#include <pwd.h>
#include <netdb.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <getopt.h>
#include <netinet/in.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#include <sys/stat.h>

#define _PATH_FINGER    "/usr/bin/finger"
#define _PATH_UPTIME    "/usr/bin/uptime"
#define	ENTRIES	50
#define WS " \t\r\n"

/* These are used in this order if the finger path compiled in doesn't work. */
#define _ALT_PATH_FINGER_1 "/usr/local/bin/finger"
#define _ALT_PATH_FINGER_2 "/usr/ucb/finger"
#define _ALT_PATH_FINGER_3 "/usr/bin/finger"

static
void
fatal(const char *msg, int logit)
{
	const char *err = strerror(errno);
	if (logit) syslog(LOG_ERR, "%s: %s:\n", msg, err);
	fprintf(stderr, "fingerd: %s: %s\r\n", msg, err);
	exit(1);
}


int
main(int argc, char *argv[])
{
	struct passwd *pw;
	FILE *fp;
	int ch, ca;
	int p[2];
	const char *av[ENTRIES + 1];
	char line[1024];
	int welcome = 0, heavylogging = 0, nouserlist = 0;
	int k, pid, nusers;
	char *s, *t;
	const char *fingerpath = NULL;
       int server = 0;
       char tmp_stuff[25];
       char filename[100];
       char cha;
       char str[2];
       struct stat statbuf;
       FILE *in;

	struct sockaddr_in sn;
	int sval = sizeof(sn);
	if (getpeername(0, (struct sockaddr *) &sn, &sval) < 0) {
		fatal("getpeername", 0);
	}

	opterr = 0;
	while ((ca = getopt(argc, argv, "wlsLuh?")) != EOF) {
		switch(ca) {
		  case 'w':
			welcome = 1;
			break;
		  case 'l':
		        heavylogging = 1;
			break;
		  case 's':
		  	server = 1;
		  	break;
		  case 'L':
		        fingerpath = optarg;
			break;
		  case 'u':
		        nouserlist = 1;
			break;
		  case '?':
		  case 'h':
		  default:
			syslog(LOG_ERR, "usage: fingerd [-lLwu]");
			exit(1);
		}
	}
	argc -= optind;
	argv += optind;

#ifdef USE_FILE_SERVER
    server = 1;
#endif


        if (((pw = getpwnam("nobody")) != NULL)&&(server != 1)) {
           setgid(pw->pw_gid);
           setuid(pw->pw_uid);
        }

	openlog("fingerd", LOG_PID, LOG_DAEMON);

        printf("Your finger attempt was logged.\r\n\n");

	if (!fgets(line, sizeof(line), stdin))
		exit(1);

	if (welcome) {
		char buf[256];
		struct hostent *hp;
		struct utsname utsname;

		uname(&utsname);
		gethostname(buf, sizeof(buf));
		if ((hp = gethostbyname(buf))) {
			/* paranoia: dns spoofing? */
			strncpy(buf, hp->h_name, sizeof(buf));
			buf[sizeof(buf)-1] = 0;
		}
		printf("\r\nWelcome to %s version %s at %s !\r\n\n",
				utsname.sysname, utsname.release, buf);
		fflush(stdout);
		switch (fork()) {
		 case -1: /* fork failed, oh well */
		     break;
		 case 0: /* child */
		     execl(_PATH_UPTIME, _PATH_UPTIME, NULL);
		     _exit(1);
		 default: /* parent */
		     wait(NULL);
		     break;
		}
		printf("\r\n");
		fflush(stdout);
	}

	k = nusers = 0;
	av[k++] = "finger";
	for (s = strtok(line, WS); s && k<ENTRIES; s = strtok(NULL, WS)) {
		/* RFC742: "/[Ww]" == "-l" */
		if (!strncasecmp(s, "/w", 2)) memcpy(s, "-l", 2);
		t = strchr(s, '@');
		if (t) {
			fprintf(stderr, "fingerd: forwarding not allowed\r\n");
			syslog(LOG_WARNING, "rejected forwarding %s\n", s);
			exit(1);
		}
		av[k++] = s;
		if (*s!='-') nusers++;
                if (heavylogging)  if (*s=='-') syslog(LOG_INFO, "option %s\n", s);
		                                      
	}
        fflush(stdout);
                                              	
	av[k] = NULL;
	if (nusers==0) {
		/* finger @host */
		if (nouserlist) {
			syslog(LOG_WARNING, "rejected finger @host\n");
			printf("Please supply a username\r\n");
			return 0;
		}
		if (heavylogging) syslog(LOG_INFO, "fingered @host\n");
	}

	if (pipe(p) < 0) fatal("pipe", 1);

	pid = fork();
	if (pid<0) fatal("fork", 1);
	if (pid==0) {
		/* child */
		close(p[0]);
		dup2(p[1], 1);
		if (p[1]!=1) close(p[1]);
    strncpy(tmp_stuff,line,sizeof(tmp_stuff));
    if (tmp_stuff[strlen(tmp_stuff)-1] == 10) tmp_stuff[strlen(tmp_stuff)-1] = 0;
    if ((tmp_stuff[0] == '/') || (strstr(tmp_stuff,"..")) != NULL) {
       printf("No tricks guy ...\r\n");
       syslog(LOG_INFO, "haxor: %s\n", tmp_stuff);
       fflush(stdout);
       _exit(1);    
    }
    strcpy(filename, FILE_DIR);
    strcat(filename, tmp_stuff);
    if (stat(filename, &statbuf) >= 0) {
       in = fopen(filename,"r");
       while ((cha = (char)getc(in)) != EOF) {
          str[0] = cha;
          str[1] = '\0';
          printf("%s",str);
       }
       fflush(stdout);
       fclose(in);
       _exit(1);
    } else {
        syslog(LOG_INFO, "fingered %s\n", tmp_stuff);
#ifdef NO_USER
    _exit(1);
#endif
	if (fingerpath) execv(fingerpath, (char **)av);
		execv(_PATH_FINGER, (char **)av);
		execv(_ALT_PATH_FINGER_1, (char **)av);
		execv(_ALT_PATH_FINGER_2, (char **)av);
		execv(_ALT_PATH_FINGER_3, (char **)av);
                _exit(1);
    }
	}
	/* parent */
	close(p[1]);

	/* convert \n to \r\n. This should be an option to finger... */
	fp = fdopen(p[0], "r");
	if (!fp) fatal("fdopen", 1);

	while ((ch = getc(fp)) != EOF) {
		if (ch == '\n')	putchar('\r');
		putchar(ch);
	}
	return 0;
}
